# Author: Yiannis Charalambous 2023

from typing import Optional
from typing_extensions import override
from langchain.base_language import BaseLanguageModel
from langchain.schema import BaseMessage

from esbmc_ai.chat_response import ChatResponse, FinishReason
from esbmc_ai.config import ChatPromptSettings, DynamicAIModelAgent
from esbmc_ai.frontend.solution import apply_line_patch

from .ai_models import AIModel
from .base_chat_interface import BaseChatInterface
from esbmc_ai.esbmc_util import (
    esbmc_get_counter_example,
    esbmc_get_violated_property,
    get_source_code_err_line_idx,
)


def get_source_code_formatted(
    source_code_format: str, source_code: str, esbmc_output: str
) -> str:
    match source_code_format:
        case "single":
            line: Optional[int] = get_source_code_err_line_idx(esbmc_output)
            assert line, "error line not found in esbmc output"
            # ESBMC reports errors starting from 1. To get the correct line, we need to use 0 based
            # indexing.
            return source_code.splitlines(True)[line]
        case "full":
            return source_code
        case _:
            raise ValueError(
                f"Not a valid format for source code: {source_code_format}"
            )


def get_esbmc_output_formatted(esbmc_output_type: str, esbmc_output: str) -> str:
    match esbmc_output_type:
        case "vp":
            value: Optional[str] = esbmc_get_violated_property(esbmc_output)
            if not value:
                raise ValueError("Not found violated property.")
            return value
        case "ce":
            value: Optional[str] = esbmc_get_counter_example(esbmc_output)
            if not value:
                raise ValueError("Not found counterexample.")
            return value
        case "full":
            return esbmc_output
        case _:
            raise ValueError(f"Not a valid ESBMC output type: {esbmc_output_type}")


class SolutionGenerator(BaseChatInterface):
    def __init__(
        self,
        ai_model_agent: DynamicAIModelAgent,
        llm: BaseLanguageModel,
        source_code: str,
        esbmc_output: str,
        ai_model: AIModel,
        scenario: str = "",
        source_code_format: str = "full",
        esbmc_output_type: str = "full",
    ) -> None:
        # Convert to chat prompt
        chat_prompt: ChatPromptSettings = DynamicAIModelAgent.to_chat_prompt_settings(
            ai_model_agent=ai_model_agent, scenario=scenario
        )

        super().__init__(
            ai_model_agent=chat_prompt,
            ai_model=ai_model,
            llm=llm,
        )
        self.initial_prompt = ai_model_agent.initial_prompt

        self.esbmc_output_type: str = esbmc_output_type
        self.esbmc_output = get_esbmc_output_formatted(
            esbmc_output_type=self.esbmc_output_type,
            esbmc_output=esbmc_output,
        )

        self.source_code_format: str = source_code_format
        self.source_code_raw: str = source_code
        self.source_code = get_source_code_formatted(
            source_code_format=self.source_code_format,
            source_code=self.source_code_raw,
            esbmc_output=self.esbmc_output,
        )

        self.set_template_value("source_code", self.source_code)
        self.set_template_value("esbmc_output", self.esbmc_output)

    @override
    def compress_message_stack(self) -> None:
        # Resets the conversation - cannot summarize code
        self.messages: list[BaseMessage] = []

    @classmethod
    def get_code_from_solution(cls, solution: str) -> str:
        """Strip the source code of any leftover text as sometimes the AI model
        will generate text and formatting despite being told not to."""
        try:
            code_start: int = solution.index("```") + 3
            assert code_start != -1

            # Remove up until the new line, because usually there's a language
            # specification after the 3 ticks ```c...
            code_start = solution.index("\n", code_start) + 1
            assert code_start != -1

            code_end: int = solution[::-1].index("```")
            assert code_start != -1

            # -4 = 3 backticks and also the \n before the backticks.
            code_end: int = len(solution) - 4 - code_end
            assert code_start <= code_end

            solution = solution[code_start:code_end]
        except (ValueError, AssertionError):
            pass
        return solution

    def generate_solution(self) -> tuple[str, FinishReason]:
        response: ChatResponse = self.send_message(self.initial_prompt)
        solution: str = str(response.message.content)

        solution = SolutionGenerator.get_code_from_solution(solution)

        # If source code passed to LLM is formatted then we need to recombine to
        # full source code before giving to ESBMC
        match self.source_code_format:
            case "single":
                err_line: Optional[int] = get_source_code_err_line_idx(
                    self.esbmc_output
                )
                assert (
                    err_line
                ), "fix code command: error line could not be found to apply brutal patch replacement"
                solution = apply_line_patch(
                    self.source_code_raw, solution, err_line, err_line
                )

        return solution, response.finish_reason



# Author: Yiannis Charalambous 2023

import os
import json
import sys
from platform import system as system_name
from pathlib import Path
from dotenv import load_dotenv, find_dotenv

from typing import Any, NamedTuple, Optional, Union, Sequence
from dataclasses import dataclass
from langchain.schema import BaseMessage

from .logging import *
from .ai_models import *
from .api_key_collection import APIKeyCollection
from .chat_response import json_to_base_messages


api_keys: APIKeyCollection

esbmc_path: str = "~/.local/bin/esbmc"
esbmc_params: list[str] = [
    "--interval-analysis",
    "--goto-unwind",
    "--unlimited-goto-unwind",
    "--k-induction",
    "--state-hashing",
    "--add-symex-value-sets",
    "--k-step",
    "2",
    "--floatbv",
    "--unlimited-k-steps",
    "--compact-trace",
    "--context-bound",
    "2",
]

temp_auto_clean: bool = True
temp_file_dir: str = "."
ai_model: AIModel = AIModels.GPT_3.value

esbmc_output_type: str = "full"
source_code_format: str = "full"

fix_code_max_attempts: int = 5

requests_max_tries: int = 5
requests_timeout: float = 60
verifier_timeout: float = 60

loading_hints: bool = False
allow_successful: bool = False
# Show the raw conversation after the command ends
raw_conversation: bool = False

cfg_path: str


class AIAgentConversation(NamedTuple):
    """Immutable class describing the conversation definition for an AI agent. The
    class represents the system messages of the AI agent defined and contains a load
    class method for efficiently loading it from config."""

    messages: tuple[BaseMessage, ...]

    @classmethod
    def from_seq(cls, message_list: Sequence[BaseMessage]) -> "AIAgentConversation":
        return cls(messages=tuple(message_list))

    @classmethod
    def load_from_config(
        cls, messages_list: list[dict[str, str]]
    ) -> "AIAgentConversation":
        return cls(messages=tuple(json_to_base_messages(messages_list)))


@dataclass
class ChatPromptSettings:
    """Settings for the AI Model. These settings act as an actor/agent, allowing the
    AI model to be applied into a specific scenario."""

    system_messages: AIAgentConversation
    """The generic prompt system messages of the AI. Generic meaning it is used in
    every scenario, as opposed to dynamic system message. The value is a list of
    converstaions."""
    initial_prompt: str
    """The generic initial prompt to use for the agent."""
    temperature: float


@dataclass
class DynamicAIModelAgent(ChatPromptSettings):
    """Extension of the ChatPromptSettings to include dynamic"""

    scenarios: dict[str, AIAgentConversation]
    """Scenarios dictionary that contains system messages for different errors that
    ESBMC can give. More information can be found in the
    [wiki](https://github.com/Yiannis128/esbmc-ai/wiki/Configuration#dynamic-prompts). 
    Reads from the config file the following hierarchy:
    * Dictionary mapping of error type to dictionary. Accepts the following entries:
      * `system` mapping to an array. The array contains the conversation for the
      system message for this particular error."""

    @classmethod
    def to_chat_prompt_settings(
        cls, ai_model_agent: "DynamicAIModelAgent", scenario: str
    ) -> ChatPromptSettings:
        """DynamicAIModelAgent extensions are not used by BaseChatInterface derived classes
        directly, since they only use the SystemMessages of ChatPromptSettings. This applies
        the correct scenario as a System Message and returns a pure ChatPromptSettings object
        for use. **Will return a shallow copy even if the system message is to be used**.
        """
        if scenario in ai_model_agent.scenarios:
            return ChatPromptSettings(
                initial_prompt=ai_model_agent.initial_prompt,
                system_messages=ai_model_agent.scenarios[scenario],
                temperature=ai_model_agent.temperature,
            )
        else:
            return ChatPromptSettings(
                initial_prompt=ai_model_agent.initial_prompt,
                system_messages=ai_model_agent.system_messages,
                temperature=ai_model_agent.temperature,
            )


chat_prompt_user_mode: DynamicAIModelAgent
chat_prompt_generator_mode: DynamicAIModelAgent
chat_prompt_optimize_code: ChatPromptSettings

esbmc_params_optimize_code: list[str] = [
    "--incremental-bmc",
    "--no-bounds-check",
    "--no-pointer-check",
    "--no-div-by-zero-check",
]


def _load_custom_ai(config: dict) -> None:
    ai_custom: dict = config
    for name, ai_data in ai_custom.items():
        # Load the max tokens
        custom_ai_max_tokens, ok = _load_config_value(
            config_file=ai_data,
            name="max_tokens",
        )
        assert ok, f'max_tokens field not found in "ai_custom" entry "{name}".'
        assert (
            isinstance(custom_ai_max_tokens, int) and custom_ai_max_tokens > 0
        ), f'custom_ai_max_tokens in ai_custom entry "{name}" needs to be an int and greater than 0.'
        # Load the URL
        custom_ai_url, ok = _load_config_value(
            config_file=ai_data,
            name="url",
        )
        assert ok, f'url field not found in "ai_custom" entry "{name}".'
        stop_sequences, ok = _load_config_value(
            config_file=ai_data,
            name="stop_sequences",
        )
        # Load the config message
        config_message: dict[str, str] = ai_data["config_message"]
        template, ok = _load_config_value(
            config_file=config_message,
            name="template",
        )
        human, ok = _load_config_value(
            config_file=config_message,
            name="human",
        )
        ai, ok = _load_config_value(
            config_file=config_message,
            name="ai",
        )
        system, ok = _load_config_value(
            config_file=config_message,
            name="system",
        )

        # Add the custom AI.
        add_custom_ai_model(
            AIModelTextGen(
                name=name,
                tokens=custom_ai_max_tokens,
                url=custom_ai_url,
                config_message=template,
                ai_template=ai,
                human_template=human,
                system_template=system,
                stop_sequences=stop_sequences,
            )
        )


def load_envs() -> None:
    """Environment variables are loaded in the following order:

    1. Environment variables already loaded. Any variable not present will be looked for in
    .env files in the following locations.
    2. .env file in the current directory, moving upwards in the directory tree.
    3. esbmc-ai.env file in the current directory, moving upwards in the directory tree.
    4. esbmc-ai.env file in $HOME/.config/ for Linux/macOS and %userprofile% for Windows.

    Note: ESBMC_AI_CFG_PATH undergoes tilde user expansion and also environment
    variable expansion.
    """

    def get_env_vars() -> None:
        """Gets all the system environment variables that are currently loaded."""
        for k in keys:
            value: Optional[str] = os.getenv(k)
            if value != None:
                values[k] = value

    keys: list[str] = ["OPENAI_API_KEY", "HUGGINGFACE_API_KEY", "ESBMC_AI_CFG_PATH"]
    values: dict[str, str] = {}

    # Load from system env
    get_env_vars()

    # Find .env in current working directory and load it.
    dotenv_file_path: str = find_dotenv(usecwd=True)
    if dotenv_file_path != "":
        load_dotenv(dotenv_path=dotenv_file_path, override=False, verbose=True)
    else:
        # Find esbmc-ai.env in current working directory and load it.
        dotenv_file_path: str = find_dotenv(filename="esbmc-ai.env", usecwd=True)
        if dotenv_file_path != "":
            load_dotenv(dotenv_path=dotenv_file_path, override=False, verbose=True)

    get_env_vars()

    # Look for .env in home folder.
    home_path: Path = Path.home()
    match system_name():
        case "Linux" | "Darwin":
            home_path /= ".config/esbmc-ai.env"
        case "Windows":
            home_path /= "esbmc-ai.env"
        case _:
            raise ValueError(f"Unknown OS type: {system_name()}")

    load_dotenv(home_path, override=False, verbose=True)

    get_env_vars()

    # Check if all the values are set, else error.
    for key in keys:
        if key not in values:
            print(f"Error: No ${key} in environment.")
            sys.exit(1)

    global api_keys
    api_keys = APIKeyCollection(
        openai=str(os.getenv("OPENAI_API_KEY")),
        huggingface=str(os.getenv("HUGGINGFACE_API_KEY")),
    )

    global cfg_path
    cfg_path = os.path.expanduser(
        os.path.expandvars(str(os.getenv("ESBMC_AI_CFG_PATH")))
    )


def _load_ai_data(config: dict) -> None:
    # User chat mode will store extra AIAgentConversations into scenarios.
    global chat_prompt_user_mode
    chat_prompt_user_mode = DynamicAIModelAgent(
        system_messages=AIAgentConversation.load_from_config(
            config["chat_modes"]["user_chat"]["system"]
        ),
        initial_prompt=config["chat_modes"]["user_chat"]["initial"],
        temperature=config["chat_modes"]["user_chat"]["temperature"],
        scenarios={
            "set_solution": AIAgentConversation.load_from_config(
                messages_list=config["chat_modes"]["user_chat"]["set_solution"],
            ),
        },
    )

    # Generator mode loads scenarios normally.
    json_fcm_scenarios: dict = config["chat_modes"]["generate_solution"]["scenarios"]
    fcm_scenarios: dict = {
        scenario: AIAgentConversation.load_from_config(messages["system"])
        for scenario, messages in json_fcm_scenarios.items()
    }
    global chat_prompt_generator_mode
    chat_prompt_generator_mode = DynamicAIModelAgent(
        system_messages=AIAgentConversation.load_from_config(
            config["chat_modes"]["generate_solution"]["system"]
        ),
        initial_prompt=config["chat_modes"]["generate_solution"]["initial"],
        temperature=config["chat_modes"]["generate_solution"]["temperature"],
        scenarios=fcm_scenarios,
    )


def _load_config_value(
    config_file: dict, name: str, default: object = None
) -> tuple[Any, bool]:
    if name in config_file:
        return config_file[name], True
    else:
        print(f"Warning: {name} not found in config... Using default value: {default}")
        return default, False


def _load_config_bool(
    config_file: dict,
    name: str,
    default: bool = False,
) -> bool:
    value, _ = _load_config_value(config_file, name, default)
    if isinstance(value, bool):
        return value
    else:
        raise TypeError(
            f"Error: config invalid {name} value: {value} "
            + "Make sure it is a bool value."
        )


def _load_config_real_number(
    config_file: dict, name: str, default: object = None
) -> Union[int, float]:
    value, _ = _load_config_value(config_file, name, default)
    # Type check
    if type(value) is float or type(value) is int:
        return value
    else:
        raise TypeError(
            f"Error: config invalid {name} value: {value} "
            + "Make sure it is a float or int..."
        )


def load_config(file_path: str) -> None:
    if not os.path.exists(file_path) and os.path.isfile(file_path):
        print(f"Error: Config not found: {file_path}")
        sys.exit(1)

    config_file = None
    with open(file_path, mode="r") as file:
        config_file = json.load(file)

    global esbmc_params
    esbmc_params, _ = _load_config_value(
        config_file,
        "esbmc_params",
        esbmc_params,
    )

    global fix_code_max_attempts
    fix_code_max_attempts = int(
        _load_config_real_number(
            config_file=config_file["chat_modes"]["generate_solution"],
            name="max_attempts",
            default=fix_code_max_attempts,
        )
    )

    global source_code_format
    source_code_format, _ = _load_config_value(
        config_file=config_file,
        name="source_code_format",
        default=source_code_format,
    )

    if source_code_format not in ["full", "single"]:
        raise Exception(
            f"Source code format in the config is not valid: {source_code_format}"
        )

    global esbmc_output_type
    esbmc_output_type, _ = _load_config_value(
        config_file=config_file,
        name="esbmc_output_type",
        default=esbmc_output_type,
    )

    if esbmc_output_type not in ["full", "vp", "ce"]:
        raise Exception(
            f"ESBMC output type in the config is not valid: {esbmc_output_type}"
        )

    global requests_max_tries
    requests_max_tries = int(
        _load_config_real_number(
            config_file=config_file["llm_requests"],
            name="max_tries",
            default=requests_max_tries,
        )
    )

    global requests_timeout
    requests_timeout = _load_config_real_number(
        config_file=config_file["llm_requests"],
        name="timeout",
        default=requests_timeout,
    )

    global verifier_timeout
    verifier_timeout = _load_config_real_number(
        config_file=config_file,
        name="verifier_timeout",
        default=verifier_timeout,
    )

    global temp_auto_clean
    temp_auto_clean, _ = _load_config_value(
        config_file,
        "temp_auto_clean",
        temp_auto_clean,
    )

    global temp_file_dir
    temp_file_dir, _ = _load_config_value(
        config_file,
        "temp_file_dir",
        temp_file_dir,
    )

    global allow_successful
    allow_successful = _load_config_bool(
        config_file,
        "allow_successful",
        False,
    )

    global loading_hints
    loading_hints = _load_config_bool(
        config_file,
        "loading_hints",
        True,
    )

    # Load the custom ai configs.
    _load_custom_ai(config_file["ai_custom"])

    global ai_model
    ai_model_name, _ = _load_config_value(
        config_file,
        "ai_model",
        ai_model,
    )
    if is_valid_ai_model(ai_model_name):
        # Load the ai_model from loaded models.
        ai_model = get_ai_model_by_name(ai_model_name)
    else:
        print(f"Error: {ai_model_name} is not a valid AI model")
        sys.exit(4)

    global esbmc_path
    # Health check verifies this later in the init process.
    esbmc_path, _ = _load_config_value(
        config_file,
        "esbmc_path",
        esbmc_path,
    )
    # Expand variables and tilde.
    esbmc_path = os.path.expanduser(os.path.expandvars(esbmc_path))

    # Load the AI data from the file that will command the AI for all modes.
    printv("Initializing AI data")
    _load_ai_data(config=config_file)


def load_args(args) -> None:
    set_verbose(args.verbose)

    global ai_model
    if args.ai_model != "":
        if is_valid_ai_model(args.ai_model):
            ai_model = get_ai_model_by_name(args.ai_model)
        else:
            print(f"Error: invalid --ai-model parameter {args.ai_model}")
            sys.exit(4)

    global raw_conversation
    raw_conversation = args.raw_conversation

    global esbmc_params
    # If append flag is set, then append.
    if args.append:
        esbmc_params.extend(args.remaining)
    elif len(args.remaining) != 0:
        esbmc_params = args.remaining



# Author: Yiannis Charalambous

import os
import sys
from subprocess import PIPE, STDOUT, run, CompletedProcess
from pathlib import Path
from typing import Optional

from . import config


def esbmc_get_violated_property(esbmc_output: str) -> Optional[str]:
    """Gets the violated property line of the ESBMC output."""
    # Find "Violated property:" string in ESBMC output
    lines: list[str] = esbmc_output.splitlines()
    for ix, line in enumerate(lines):
        if "Violated property:" == line:
            return "\n".join(lines[ix : ix + 3])
    return None


def esbmc_get_counter_example(esbmc_output: str) -> Optional[str]:
    """Gets ESBMC output after and including [Counterexample]"""
    idx: int = esbmc_output.find("[Counterexample]\n")
    if idx == -1:
        return None
    else:
        return esbmc_output[idx:]


def esbmc_get_error_type(esbmc_output: str) -> str:
    """Gets the error of violated property, the entire line."""
    # TODO Test me
    # Start search from the marker.
    marker: str = "Violated property:\n"
    violated_property_index: int = esbmc_output.rfind(marker) + len(marker)
    from_loc_error_msg: str = esbmc_output[violated_property_index:]
    # Find second new line which contains the location of the violated
    # property and that should point to the line with the type of error.
    # In this case, the type of error is the "scenario".
    scenario_index: int = from_loc_error_msg.find("\n")
    scenario: str = from_loc_error_msg[scenario_index + 1 :]
    scenario_end_l_index: int = scenario.find("\n")
    scenario = scenario[:scenario_end_l_index].strip()
    return scenario


def get_source_code_err_line(esbmc_output: str) -> Optional[int]:
    # Find "Violated property:" string in ESBMC output
    violated_property: Optional[str] = esbmc_get_violated_property(esbmc_output)
    if violated_property:
        # Get the line of the violated property.
        pos_line: str = violated_property.splitlines()[1]
        pos_line_split: list[str] = pos_line.split(" ")
        for ix, word in enumerate(pos_line_split):
            if word == "line":
                # Get the line number
                return int(pos_line_split[ix + 1])
    return None


def get_source_code_err_line_idx(esbmc_output: str) -> Optional[int]:
    line: Optional[int] = get_source_code_err_line(esbmc_output)
    if line:
        return line - 1
    else:
        return None


def esbmc(path: str, esbmc_params: list, timeout: Optional[float] = None):
    """Exit code will be 0 if verification successful, 1 if verification
    failed. And any other number for compilation error/general errors."""
    # Build parameters
    esbmc_cmd = [config.esbmc_path]
    esbmc_cmd.extend(esbmc_params)
    esbmc_cmd.append(path)

    if "--timeout" in esbmc_cmd:
        print(
            'Do not add --timeout to ESBMC parameters, instead specify it in "verifier_timeout".'
        )
        sys.exit(1)

    esbmc_cmd.extend(["--timeout", str(timeout)])

    # Add slack time to process to allow verifier to timeout and end gracefully.
    process_timeout: Optional[float] = timeout + 1 if timeout else None

    # Run ESBMC and get output
    process: CompletedProcess = run(
        esbmc_cmd,
        stdout=PIPE,
        stderr=STDOUT,
        timeout=process_timeout,
    )

    output: str = process.stdout.decode("utf-8")
    return process.returncode, output


def esbmc_load_source_code(
    file_path: str,
    source_code: str,
    esbmc_params: list = config.esbmc_params,
    auto_clean: bool = config.temp_auto_clean,
    timeout: Optional[float] = None,
):
    source_code_path = Path(file_path)

    # Create temp path.
    delete_path: bool = False
    if not os.path.exists(config.temp_file_dir):
        os.mkdir(config.temp_file_dir)
        delete_path = True

    temp_file_path = f"{config.temp_file_dir}{os.sep}{source_code_path.name}"

    # Create temp file.
    with open(temp_file_path, "w") as file:
        # Save to temporary folder and flush contents.
        file.write(source_code)
        file.flush()

        # Call ESBMC to temporary folder.
        results = esbmc(file.name, esbmc_params, timeout=timeout)

    # Delete temp files and path
    if auto_clean:
        # Remove file
        os.remove(temp_file_path)
        # Remove file path if created this run and is empty.
        if delete_path and len(os.listdir(config.temp_file_dir)) == 0:
            os.rmdir(config.temp_file_dir)

    # Return
    return results



# Author: Yiannis Charalambous

from langchain.prompts.chat import ChatPromptValue
from langchain.schema import (
    AIMessage,
    HumanMessage,
    PromptValue,
    SystemMessage,
)
from pytest import raises

from esbmc_ai.ai_models import (
    add_custom_ai_model,
    is_valid_ai_model,
    AIModel,
    AIModels,
    get_ai_model_by_name,
    AIModelTextGen,
)


def test_is_valid_ai_model() -> None:
    assert is_valid_ai_model(AIModels.FALCON_7B.value)
    assert is_valid_ai_model(AIModels.GPT_3_16K.value)
    assert is_valid_ai_model("gpt-3.5-turbo")
    assert is_valid_ai_model("falcon-7b")


def test_is_not_valid_ai_model() -> None:
    custom_model: AIModel = AIModel(
        name="custom_ai",
        tokens=999,
    )
    assert not is_valid_ai_model(custom_model)
    assert not is_valid_ai_model("doesn't exist")


def test_add_custom_ai_model() -> None:
    custom_model: AIModel = AIModel(
        name="custom_ai",
        tokens=999,
    )

    add_custom_ai_model(custom_model)

    assert is_valid_ai_model(custom_model.name)

    # Test add again.

    if is_valid_ai_model(custom_model.name):
        with raises(Exception):
            add_custom_ai_model(custom_model)

    assert is_valid_ai_model(custom_model.name)


def test_get_ai_model_by_name() -> None:
    # Try with first class AI
    assert get_ai_model_by_name("gpt-3.5-turbo")

    # Try with custom AI.
    # Add custom AI model if not added by previous tests.
    custom_model: AIModel = AIModel(
        name="custom_ai",
        tokens=999,
    )
    if not is_valid_ai_model(custom_model.name):
        add_custom_ai_model(custom_model)
    assert get_ai_model_by_name("custom_ai") is not None

    # Try with non existent.
    with raises(Exception):
        get_ai_model_by_name("not-exists")


def test_apply_chat_template() -> None:
    messages: list = [
        SystemMessage(content="M1"),
        HumanMessage(content="M2"),
        AIMessage(content="M3"),
    ]

    # Test the identity method.
    custom_model_1: AIModel = AIModel(
        name="custom",
        tokens=999,
    )

    prompt: PromptValue = custom_model_1.apply_chat_template(messages=messages)

    assert prompt == ChatPromptValue(messages=messages)

    # Test the text gen method
    custom_model_2: AIModelTextGen = AIModelTextGen(
        name="custom",
        tokens=999,
        url="",
        config_message="{history}\n\n{user_prompt}",
        ai_template="AI: {content}",
        human_template="Human: {content}",
        system_template="System: {content}",
    )

    prompt_text: str = custom_model_2.apply_chat_template(messages=messages).to_string()

    assert prompt_text == "System: M1\n\nHuman: M2\n\nAI: M3"


def test_escape_messages() -> None:
    """Tests that the brackets are escaped properly using `AIModel.escape_message`."""

    messages = [
        HumanMessage(content="Hello my name is {name} and I like {{apples}}"),
        SystemMessage(content="Hello my {{name is system} and I like {{{apples}}}"),
        AIMessage(content="Hello my {{ {{{apples}}"),
        SystemMessage(content="{descreption}{descreption}{{descreption}}{descreption}"),
        SystemMessage(content="{descreption}{{{descreption}}}{{{{{descreption}}}}}"),
        SystemMessage(
            content="{apples}{{{apples}}}{{{{{apples}}}}}{{{{{{{apples}}}}}}}"
        ),
    ]

    allowed = ["name", "descreption"]

    filtered = [
        HumanMessage(content="Hello my name is {name} and I like {{apples}}"),
        SystemMessage(content="Hello my {{name is system}} and I like {{{{apples}}}}"),
        AIMessage(content="Hello my {{ {{{{apples}}"),
        SystemMessage(content="{descreption}{descreption}{{descreption}}{descreption}"),
        SystemMessage(content="{descreption}{{{descreption}}}{{{{{descreption}}}}}"),
        SystemMessage(
            content="{{apples}}{{{{apples}}}}{{{{{{apples}}}}}}{{{{{{{{apples}}}}}}}}"
        ),
    ]

    result = list(AIModel.escape_messages(messages, allowed))

    assert result[0] == filtered[0]
    assert result[1] == filtered[1]
    assert result[2] == filtered[2]
    assert result[3] == filtered[3]
    assert result[4] == filtered[4]
    assert result[5] == filtered[5]



# Author: Yiannis Charalambous

"""# Solution
Keeps track of all the source files that ESBMC-AI is targeting."""

from typing import NamedTuple


class SourceFile(NamedTuple):
    file_path: str
    content: str


_main_source_file: SourceFile = SourceFile("", "")
_source_files: set[SourceFile] = set()


def add_source_file(source_file: SourceFile) -> None:
    global _source_files
    _source_files.add(source_file)


def set_main_source_file(source_file: SourceFile) -> None:
    add_source_file(source_file)
    global _main_source_file
    _main_source_file = source_file


def get_main_source_file_path() -> str:
    global _main_source_file
    return _main_source_file.file_path


def get_main_source_file() -> SourceFile:
    global _main_source_file
    return _main_source_file


def get_source_files() -> list[SourceFile]:
    global _source_files
    return list(_source_files)


def apply_line_patch(source_code: str, patch: str, start: int, end: int) -> str:
    """Applies a patch to the source code.

    To replace a single line, start and end are equal.

    Args:
        * source_code - The source code to apply the patch to.
        * patch - Can be a line or multiple lines but will replace the start and
        end region defined.
        * start - Line index to mark start of replacement.
        * end - Marks the end of the region where the patch will be applied to.
        End is non-inclusive."""
    assert (
        start <= end
    ), f"start ({start}) needs to be less than or equal to end ({end})"
    lines: list[str] = source_code.splitlines()
    lines = lines[:start] + [patch] + lines[end + 1 :]
    return "\n".join(lines)



# Author: Yiannis Charalambous 2023

from pytest import raises

import esbmc_ai.config as config

from esbmc_ai.ai_models import is_valid_ai_model


def test_load_config_value() -> None:
    result, ok = config._load_config_value(
        {
            "test": "value",
        },
        "test",
    )
    assert ok and result == "value"


def test_load_config_value_default_value() -> None:
    result, ok = config._load_config_value(
        {
            "test": "value",
        },
        "test",
        "wrong",
    )
    assert ok and result == "value"


def test_load_config_value_default_value_not_exists() -> None:
    result, ok = config._load_config_value(
        {},
        "test2",
        "wrong",
    )
    assert not ok and result == "wrong"


def test_load_config_real_number() -> None:
    result = config._load_config_real_number(
        {
            "test": 1.0,
        },
        "test",
    )
    assert result == 1.0


def test_load_config_real_number_default_value() -> None:
    result = config._load_config_real_number({}, "test", 1.1)
    assert result == 1.1


def test_load_config_real_number_wrong_value() -> None:
    with raises(TypeError):
        result = config._load_config_real_number(
            {
                "test": "wrong value",
            },
            "test",
        )
        assert result == None


def test_load_config_real_number_wrong_value_default() -> None:
    with raises(TypeError):
        result = config._load_config_real_number(
            {
                "test": "wrong value",
            },
            "test",
            1.0,
        )
        assert result == None


def test_load_custom_ai() -> None:
    custom_ai_config: dict = {
        "example_ai": {
            "max_tokens": 4096,
            "url": "www.example.com",
            "config_message": {
                "template": "example",
                "system": "{content}",
                "ai": "{content}",
                "human": "{content}",
            },
        }
    }

    config._load_custom_ai(custom_ai_config)

    assert is_valid_ai_model("example_ai")


def test_load_custom_ai_fail() -> None:
    # Wrong max_tokens type
    ai_conf: dict = {
        "example_ai_2": {
            "max_tokens": "1024",
            "url": "www.example.com",
            "config_message": "example",
        }
    }

    with raises(AssertionError):
        config._load_custom_ai(ai_conf)

    # Wrong max_tokens value
    ai_conf: dict = {
        "example_ai_2": {
            "max_tokens": 0,
            "url": "www.example.com",
            "config_message": "example",
        }
    }

    with raises(AssertionError):
        config._load_custom_ai(ai_conf)

    # Missing max_tokens
    ai_conf: dict = {
        "example_ai_2": {
            "url": "www.example.com",
            "config_message": "example",
        }
    }

    with raises(AssertionError):
        config._load_custom_ai(ai_conf)

    # Missing url
    ai_conf: dict = {
        "example_ai_2": {
            "max_tokens": 0,
            "config_message": "example",
        }
    }

    with raises(AssertionError):
        config._load_custom_ai(ai_conf)

    # Missing config message
    ai_conf: dict = {
        "example_ai_2": {
            "max_tokens": 0,
            "url": "www.example.com",
        }
    }

    with raises(AssertionError):
        config._load_custom_ai(ai_conf)

    # Test load empty
    ai_conf: dict = {}

    config._load_custom_ai(ai_conf)



# Author: Yiannis Charalambous 2023

from typing_extensions import override

from langchain.base_language import BaseLanguageModel
from langchain.memory import ConversationSummaryMemory, ChatMessageHistory
from langchain.schema import BaseMessage, PromptValue, SystemMessage

from esbmc_ai.config import AIAgentConversation, ChatPromptSettings

from .ai_models import AIModel
from .base_chat_interface import BaseChatInterface


class UserChat(BaseChatInterface):
    solution: str = ""

    def __init__(
        self,
        ai_model_agent: ChatPromptSettings,
        ai_model: AIModel,
        llm: BaseLanguageModel,
        source_code: str,
        esbmc_output: str,
        set_solution_messages: AIAgentConversation,
    ) -> None:
        super().__init__(
            ai_model_agent=ai_model_agent,
            ai_model=ai_model,
            llm=llm,
        )

        # Store source code and esbmc output in order to substitute it into the message stack.
        self.source_code: str = source_code
        self.esbmc_output: str = esbmc_output
        # The messsages for setting a new solution to the source code.
        self.set_solution_messages = set_solution_messages

        self.set_template_value("source_code", self.source_code)
        self.set_template_value("esbmc_output", self.esbmc_output)

    def set_solution(self, source_code: str) -> None:
        """Sets the solution to the problem ESBMC reported, this will inform the AI."""

        self.set_template_value("source_code_solution", source_code)

        message_prompts: PromptValue = self.ai_model.apply_chat_template(
            messages=self.set_solution_messages.messages,
            **self.template_values,
        )

        for message in message_prompts.to_messages():
            self.push_to_message_stack(message)

    def set_optimized_solution(self, source_code: str) -> None:
        # NOTE Here we use the same system message as `set_solution`.
        self.set_template_value("source_code_solution", source_code)

        message_prompts: PromptValue = self.ai_model.apply_chat_template(
            messages=self.set_solution_messages.messages,
            **self.template_values,
        )

        for message in message_prompts.to_messages():
            self.push_to_message_stack(message)

    @override
    def compress_message_stack(self) -> None:
        """Uses ConversationSummaryMemory from Langchain to summarize the conversation of all the non-protected
        messages into one summary message which is added into the conversation as a SystemMessage.
        """

        memory: ConversationSummaryMemory = ConversationSummaryMemory.from_messages(
            llm=self.llm,
            chat_memory=ChatMessageHistory(messages=self.messages),
        )

        self.messages: list[BaseMessage] = []

        self.push_to_message_stack(SystemMessage(content=memory.buffer))



from esbmc_ai.__main__ import parse_command


def test_parse() -> None:
    sentence = 'Your sentence goes "here and \\"here\\" as well."'
    result = parse_command(sentence)
    assert result == (
        "Your",
        [
            "sentence",
            "goes",
            '"here and \\"here\\" as well."',
        ],
    )


def test_parse_command() -> None:
    result = parse_command("/fix-code")
    assert result == ("/fix-code", [])


def test_parse_command_args() -> None:
    result = parse_command("/optimize-code main")
    assert result == ("/optimize-code", ["main"])



# Author: Yiannis Charalambous

from langchain.base_language import BaseLanguageModel
from langchain.schema import AIMessage, BaseMessage, HumanMessage

from esbmc_ai.config import ChatPromptSettings
from .base_chat_interface import BaseChatInterface, ChatResponse
from .ai_models import AIModel


class OptimizeCode(BaseChatInterface):
    initial_message: str

    def __init__(
        self,
        ai_model_agent: ChatPromptSettings,
        initial_message: str,
        ai_model: AIModel,
        llm: BaseLanguageModel,
    ) -> None:
        super().__init__(ai_model_agent=ai_model_agent, ai_model=ai_model, llm=llm)

        self.initial_message = initial_message

    def optimize_function(self, source_code: str, function_name: str) -> ChatResponse:
        self.messages = []
        self.push_to_message_stack(
            HumanMessage(
                content=f"Reply OK if you understand the following is the source code to optimize:\n\n{source_code}"
            )
        )
        self.push_to_message_stack(AIMessage(content="OK."))

        expanded_initial_message: str = self.initial_message.replace(
            "%s", function_name
        )
        return self.send_message(expanded_initial_message)



# Author: Yiannis Charalambous

"""Logging module for verbose printing."""

verbose: int = 0


def set_verbose(level: int) -> None:
    """Sets the verbosity level."""
    global verbose
    verbose = level


def printv(m) -> None:
    """Level 1 verbose printing."""
    if verbose > 0:
        print(m)


def printvv(m) -> None:
    """Level 2 verbose printing."""
    if verbose > 1:
        print(m)


def printvvv(m) -> None:
    """Level 3 verbose printing."""
    if verbose > 2:
        print(m)



# Author: Yiannis Charalambous

import pytest

from langchain.llms.fake import FakeListLLM
from langchain.schema import AIMessage, BaseMessage, HumanMessage, SystemMessage
from esbmc_ai.ai_models import AIModel
from esbmc_ai.base_chat_interface import BaseChatInterface
from esbmc_ai.chat_response import ChatResponse
from esbmc_ai.config import AIAgentConversation, ChatPromptSettings


@pytest.fixture(scope="module")
def setup():
    ai_model: AIModel = AIModel("test", 1024)

    system_messages: list[BaseMessage] = [
        SystemMessage(content="First system message"),
        AIMessage(content="OK"),
    ]

    return ai_model, system_messages


def test_push_message_stack(setup) -> None:
    llm: FakeListLLM = FakeListLLM(responses=[])

    ai_model, system_messages = setup

    chat: BaseChatInterface = BaseChatInterface(
        ai_model_agent=ChatPromptSettings(
            AIAgentConversation.from_seq(system_messages),
            initial_prompt="",
            temperature=1.0,
        ),
        ai_model=ai_model,
        llm=llm,
    )

    assert chat.ai_model_agent.system_messages.messages == tuple(system_messages)

    messages: list[BaseMessage] = [
        AIMessage(content="Test 1"),
        HumanMessage(content="Test 2"),
        SystemMessage(content="Test 3"),
    ]

    chat.push_to_message_stack(message=messages[0])
    chat.push_to_message_stack(message=messages[1])
    chat.push_to_message_stack(message=messages[2])

    assert chat.messages[0] == messages[0]
    assert chat.messages[1] == messages[1]
    assert chat.messages[2] == messages[2]


def test_send_message(setup) -> None:
    responses: list[str] = ["OK 1", "OK 2", "OK 3"]
    llm: FakeListLLM = FakeListLLM(responses=responses)

    ai_model, system_messages = setup

    chat: BaseChatInterface = BaseChatInterface(
        ai_model_agent=ChatPromptSettings(
            AIAgentConversation.from_seq(system_messages),
            initial_prompt="",
            temperature=1.0,
        ),
        ai_model=ai_model,
        llm=llm,
    )

    chat_responses: list[ChatResponse] = [
        chat.send_message("Test 1"),
        chat.send_message("Test 2"),
        chat.send_message("Test 3"),
    ]

    assert chat_responses[0].message.content == responses[0]
    assert chat_responses[1].message.content == responses[1]
    assert chat_responses[2].message.content == responses[2]



# Author: Yiannis Charalambous

from esbmc_ai.solution_generator import SolutionGenerator


def test_get_code_from_solution():
    assert (
        SolutionGenerator.get_code_from_solution(
            "This is a code block:\n\n```c\naaa\n```"
        )
        == "aaa"
    )
    assert (
        SolutionGenerator.get_code_from_solution(
            "This is a code block:\n\n```\nabc\n```"
        )
        == "abc"
    )
    assert (
        SolutionGenerator.get_code_from_solution("This is a code block:```abc\n```")
        == "This is a code block:```abc\n```"
    )



# Author: Yiannis Charalambous

import pytest
from os import listdir

from esbmc_ai.esbmc_util import (
    esbmc_get_counter_example,
    esbmc_get_violated_property,
    get_source_code_err_line,
)


@pytest.fixture(scope="module")
def setup_get_data() -> dict[str, str]:
    data_esbmc_output: dict[str, str] = {}

    for file_name in listdir("./tests/samples/esbmc_output/line_test/"):
        path: str = f"./tests/samples/esbmc_output/line_test/{file_name}"
        with open(path, "r") as file:
            data_esbmc_output[file_name] = file.read()

    return data_esbmc_output


def test_get_source_code_err_line(setup_get_data):
    data_esbmc_output: dict[str, str] = setup_get_data

    esbmc_output: str = data_esbmc_output["cartpole_48_safe.c-amalgamation-6.c"]
    assert get_source_code_err_line(esbmc_output) == 323

    esbmc_output = data_esbmc_output["cartpole_92_safe.c-amalgamation-14.c"]
    assert get_source_code_err_line(esbmc_output) == 221

    esbmc_output = data_esbmc_output["cartpole_95_safe.c-amalgamation-80.c"]
    assert get_source_code_err_line(esbmc_output) == 285

    esbmc_output = data_esbmc_output["cartpole_26_safe.c-amalgamation-74.c"]
    assert get_source_code_err_line(esbmc_output) == 299

    esbmc_output = data_esbmc_output["robot_5_safe.c-amalgamation-13.c"]
    assert get_source_code_err_line(esbmc_output) == 350

    esbmc_output = data_esbmc_output["vdp_1_safe.c-amalgamation-28.c"]
    assert get_source_code_err_line(esbmc_output) == 247


def test_esbmc_get_counter_example(setup_get_data) -> None:
    data_esbmc_output: dict[str, str] = setup_get_data

    esbmc_output: str = data_esbmc_output["cartpole_48_safe.c-amalgamation-6.c"]
    ce_idx: int = esbmc_output.find("[Counterexample]")
    assert esbmc_get_counter_example(esbmc_output) == esbmc_output[ce_idx:]

    esbmc_output = data_esbmc_output["cartpole_92_safe.c-amalgamation-14.c"]
    ce_idx = esbmc_output.find("[Counterexample]")
    assert esbmc_get_counter_example(esbmc_output) == esbmc_output[ce_idx:]

    esbmc_output = data_esbmc_output["cartpole_95_safe.c-amalgamation-80.c"]
    ce_idx = esbmc_output.find("[Counterexample]")
    assert esbmc_get_counter_example(esbmc_output) == esbmc_output[ce_idx:]

    esbmc_output = data_esbmc_output["cartpole_26_safe.c-amalgamation-74.c"]
    ce_idx = esbmc_output.find("[Counterexample]")
    assert esbmc_get_counter_example(esbmc_output) == esbmc_output[ce_idx:]

    esbmc_output = data_esbmc_output["robot_5_safe.c-amalgamation-13.c"]
    ce_idx = esbmc_output.find("[Counterexample]")
    assert esbmc_get_counter_example(esbmc_output) == esbmc_output[ce_idx:]

    esbmc_output = data_esbmc_output["vdp_1_safe.c-amalgamation-28.c"]
    ce_idx = esbmc_output.find("[Counterexample]")
    assert esbmc_get_counter_example(esbmc_output) == esbmc_output[ce_idx:]


def test_esbmc_get_violated_property(setup_get_data) -> None:
    data_esbmc_output: dict[str, str] = setup_get_data

    esbmc_output: str = data_esbmc_output["cartpole_48_safe.c-amalgamation-6.c"]
    start_idx: int = esbmc_output.find("Violated property:")
    end_idx: int = esbmc_output.find("VERIFICATION FAILED") - 3
    assert esbmc_get_violated_property(esbmc_output) == esbmc_output[start_idx:end_idx]

    esbmc_output = data_esbmc_output["cartpole_92_safe.c-amalgamation-14.c"]
    start_idx = esbmc_output.find("Violated property:")
    end_idx = esbmc_output.find("VERIFICATION FAILED") - 3
    assert esbmc_get_violated_property(esbmc_output) == esbmc_output[start_idx:end_idx]

    esbmc_output = data_esbmc_output["cartpole_95_safe.c-amalgamation-80.c"]
    start_idx = esbmc_output.find("Violated property:")
    end_idx = esbmc_output.find("VERIFICATION FAILED") - 3
    assert esbmc_get_violated_property(esbmc_output) == esbmc_output[start_idx:end_idx]

    esbmc_output = data_esbmc_output["cartpole_26_safe.c-amalgamation-74.c"]
    start_idx = esbmc_output.find("Violated property:")
    end_idx = esbmc_output.find("VERIFICATION FAILED") - 3
    assert esbmc_get_violated_property(esbmc_output) == esbmc_output[start_idx:end_idx]

    esbmc_output = data_esbmc_output["robot_5_safe.c-amalgamation-13.c"]
    start_idx = esbmc_output.find("Violated property:")
    end_idx = esbmc_output.find("VERIFICATION FAILED") - 3
    assert esbmc_get_violated_property(esbmc_output) == esbmc_output[start_idx:end_idx]

    esbmc_output = data_esbmc_output["vdp_1_safe.c-amalgamation-28.c"]
    start_idx = esbmc_output.find("Violated property:")
    end_idx = esbmc_output.find("VERIFICATION FAILED") - 3
    assert esbmc_get_violated_property(esbmc_output) == esbmc_output[start_idx:end_idx]



# Author: Yiannis Charalambous

from os import get_terminal_size
from typing import Any, Tuple
from typing_extensions import override
from langchain.schema import AIMessage, HumanMessage

from esbmc_ai.chat_response import FinishReason

from .chat_command import ChatCommand
from .. import config
from ..msg_bus import Signal
from ..loading_widget import create_loading_widget
from ..esbmc_util import (
    esbmc_get_error_type,
    esbmc_load_source_code,
)
from ..solution_generator import SolutionGenerator, get_esbmc_output_formatted
from ..logging import printv, printvv

# TODO Remove built in messages and move them to config.


class FixCodeCommand(ChatCommand):
    on_solution_signal: Signal = Signal()

    def __init__(self) -> None:
        super().__init__(
            command_name="fix-code",
            help_message="Generates a solution for this code, and reevaluates it with ESBMC.",
        )
        self.anim = create_loading_widget()

    @override
    def execute(self, **kwargs: Any) -> Tuple[bool, str]:
        file_name: str = kwargs["file_name"]
        source_code: str = kwargs["source_code"]
        esbmc_output: str = kwargs["esbmc_output"]

        # Parse the esbmc output here and determine what "Scenario" to use.
        scenario: str = esbmc_get_error_type(esbmc_output)

        printv(f"Scenario: {scenario}")
        printv(
            f"Using dynamic prompt..."
            if scenario in config.chat_prompt_generator_mode.scenarios
            else "Using generic prompt..."
        )

        solution_generator = SolutionGenerator(
            ai_model_agent=config.chat_prompt_generator_mode,
            source_code=source_code,
            esbmc_output=esbmc_output,
            ai_model=config.ai_model,
            llm=config.ai_model.create_llm(
                api_keys=config.api_keys,
                temperature=config.chat_prompt_generator_mode.temperature,
                requests_max_tries=config.requests_max_tries,
                requests_timeout=config.requests_timeout,
            ),
            scenario=scenario,
            source_code_format=config.source_code_format,
            esbmc_output_type=config.esbmc_output_type,
        )

        print()

        max_retries: int = config.fix_code_max_attempts
        for idx in range(max_retries):
            # Get a response. Use while loop to account for if the message stack
            # gets full, then need to compress and retry.
            llm_solution: str = ""
            while True:
                # Generate AI solution
                self.anim.start("Generating Solution... Please Wait")
                llm_solution, finish_reason = solution_generator.generate_solution()
                self.anim.stop()
                if finish_reason == FinishReason.length:
                    self.anim.start("Compressing message stack... Please Wait")
                    solution_generator.compress_message_stack()
                    self.anim.stop()
                else:
                    break

            # Print verbose lvl 2
            printvv("\nGeneration:")
            printvv("-" * get_terminal_size().columns)
            printvv(llm_solution)
            printvv("-" * get_terminal_size().columns)
            printvv("")

            # Pass to ESBMC, a workaround is used where the file is saved
            # to a temporary location since ESBMC needs it in file format.
            self.anim.start("Verifying with ESBMC... Please Wait")
            exit_code, esbmc_output = esbmc_load_source_code(
                file_path=file_name,
                source_code=llm_solution,
                esbmc_params=config.esbmc_params,
                auto_clean=config.temp_auto_clean,
                timeout=config.verifier_timeout,
            )
            self.anim.stop()

            # TODO Move this process into Solution Generator since have (beginning) is done
            # inside, and the other half is done here.
            try:
                esbmc_output = get_esbmc_output_formatted(
                    esbmc_output_type=config.esbmc_output_type,
                    esbmc_output=esbmc_output,
                )
            except ValueError:
                # Probably did not compile and so ESBMC source code is clang output.
                pass

            # Print verbose lvl 2
            printvv("-" * get_terminal_size().columns)
            printvv(esbmc_output)
            printvv("-" * get_terminal_size().columns)

            if exit_code == 0:
                self.on_solution_signal.emit(llm_solution)

                if config.raw_conversation:
                    messages: list[str] = [
                        f"{msg.type}: {msg.content}"
                        for msg in solution_generator.messages
                    ]
                    print("\n" + "\n\n".join(messages))

                return False, llm_solution

            # Failure case
            print(f"Failure {idx+1}/{max_retries}: Retrying...")
            # If final iteration no need to sleep.
            if idx < max_retries - 1:

                # Inform solution generator chat about the ESBMC response.
                # TODO Add option to customize in config.
                if exit_code != 1:
                    # The program did not compile.
                    solution_generator.push_to_message_stack(
                        message=HumanMessage(
                            content=f"Here is the ESBMC output:\n\n```\n{esbmc_output}\n```"
                        )
                    )
                else:
                    solution_generator.push_to_message_stack(
                        message=HumanMessage(
                            content=f"Here is the ESBMC output:\n\n```\n{esbmc_output}\n```"
                        )
                    )

                solution_generator.push_to_message_stack(
                    AIMessage(content="Understood.")
                )

        if config.raw_conversation:
            messages: list[str] = [
                f"{msg.type}: {msg.content}" for msg in solution_generator.messages
            ]
            print("\n" + "\n\n".join(messages))
        return True, "Failed all attempts..."



# Author: Yiannis Charalambous

from abc import ABC, abstractmethod
from typing import Any, Optional


class ChatCommand(ABC):
    command_name: str
    help_message: str

    def __init__(
        self,
        command_name: str = "",
        help_message: str = "",
    ) -> None:
        super().__init__()
        self.command_name = command_name
        self.help_message = help_message


    @abstractmethod
    def execute(self, **kwargs: Optional[Any]) -> Optional[Any]:
        raise NotImplementedError(f"Command {self.command_name} is not implemented.")



# Author: Yiannis Charalambous

import pytest

from langchain.llms.fake import FakeListLLM
from langchain.schema import AIMessage, SystemMessage

from esbmc_ai.ai_models import AIModel
from esbmc_ai.chat_response import ChatResponse, FinishReason
from esbmc_ai.config import AIAgentConversation, ChatPromptSettings
from esbmc_ai.user_chat import UserChat


@pytest.fixture
def setup():
    system_messages: list = [
        SystemMessage(content="This is a system message"),
        AIMessage(content="OK"),
    ]

    set_solution_messages = [
        SystemMessage(content="Corrected output"),
    ]

    summary_text = "THIS IS A SUMMARY OF THE CONVERSATION"
    chat: UserChat = UserChat(
        ai_model_agent=ChatPromptSettings(
            system_messages=AIAgentConversation.from_seq(system_messages),
            initial_prompt="This is initial prompt",
            temperature=1.0,
        ),
        ai_model=AIModel(name="test", tokens=12),
        llm=FakeListLLM(responses=[summary_text]),
        source_code="This is source code",
        esbmc_output="This is esbmc output",
        set_solution_messages=AIAgentConversation.from_seq(set_solution_messages),
    )

    return chat, summary_text, system_messages


def test_compress_message_stack(setup) -> None:
    chat, summary_text, system_messages = setup

    chat.messages = [SystemMessage(content=chat.ai_model_agent.initial_prompt)]

    chat.compress_message_stack()

    # Check system messages
    assert chat.ai_model_agent.system_messages.messages == tuple(system_messages)

    # Check normal messages
    assert chat.messages == [SystemMessage(content=summary_text)]


def test_automatic_compress(setup) -> None:
    chat, summary_text, system_messages = setup

    # Make the prompt extra large.
    big_prompt: str = chat.ai_model_agent.initial_prompt * 10

    response: ChatResponse = chat.send_message(big_prompt)

    assert response.finish_reason == FinishReason.length

    chat.compress_message_stack()

    # Check system messages
    assert chat.ai_model_agent.system_messages.messages == tuple(system_messages)

    # Check normal messages - Should be summarized automatically
    assert chat.messages == [SystemMessage(content=summary_text)]



# Author: Yiannis Charalambous

"""API Key Collection definition."""

from typing import NamedTuple


class APIKeyCollection(NamedTuple):
    """Class that is used to pass keys to AIModels."""

    openai: str = ""
    huggingface: str = ""



# Author: Yiannis Charalambous 2023

"""
  This program is designed to create
  and animate a simple loading animation.
  Original source: https://gist.github.com/rudrathegreat/b11daed176c8119dcedbb6b06c953590
"""

from sys import stdout as terminal
from time import sleep
from itertools import cycle
from threading import Thread
from typing import Optional

from esbmc_ai import config


class LoadingWidget(object):
    done: bool = False
    thread: Optional[Thread]
    loading_text: str
    animation: list[str]
    anim_speed: float
    anim_clear_length: int

    def __init__(
        self,
        anim_speed: float = 0.1,
        animation: list[str] = ["|", "/", "-", "\\"],
    ) -> None:
        super().__init__()
        self.anim_speed = anim_speed
        self.animation = animation

        # Find the largest animatioon
        self.anim_clear_length = 0
        for frame in self.animation:
            if len(frame) > self.anim_clear_length:
                self.anim_clear_length = len(frame)

    def _animate(self) -> None:
        for c in cycle(self.animation):
            if self.done:
                break
            # Calculate how much extra space to clear after c.
            extra_space_clear: int = self.anim_clear_length - len(c)
            terminal.write(f"\r{self.loading_text} " + c + " " * extra_space_clear)
            terminal.flush()
            sleep(self.anim_speed)

        # +1 for space between loading text and c.
        clear_length: int = len(self.loading_text) + 1 + self.anim_clear_length
        terminal.write("\r" + " " * clear_length)
        terminal.flush()
        terminal.write("\r")
        terminal.flush()

    def start(self, text: str = "Please Wait") -> None:
        if not config.loading_hints:
            return
        self.done = False
        self.loading_text = text
        self.thread = Thread(target=self._animate)
        self.thread.daemon = True
        self.thread.start()

    def stop(self) -> None:
        if not config.loading_hints:
            return
        self.done = True
        # Block until end.
        if self.thread:
            self.thread.join()


_widgets: list[LoadingWidget] = []


def create_loading_widget(
    anim_speed: float = 0.1,
    animation: list[str] = ["|", "/", "-", "\\"],
) -> LoadingWidget:
    w = LoadingWidget(anim_speed=anim_speed, animation=animation)
    _widgets.append(w)
    return w


def stop_all() -> None:
    for w in _widgets:
        w.stop()



# Author: Yiannis Charalambous

from abc import abstractmethod

from langchain.base_language import BaseLanguageModel
from langchain.schema import (
    AIMessage,
    BaseMessage,
    HumanMessage,
    LLMResult,
    PromptValue,
)

from esbmc_ai.config import ChatPromptSettings
from .chat_response import ChatResponse, FinishReason
from .ai_models import AIModel


class BaseChatInterface(object):
    def __init__(
        self,
        ai_model_agent: ChatPromptSettings,
        llm: BaseLanguageModel,
        ai_model: AIModel,
    ) -> None:
        super().__init__()
        self.ai_model: AIModel = ai_model
        self.ai_model_agent: ChatPromptSettings = ai_model_agent
        self.messages: list[BaseMessage] = []
        self.llm: BaseLanguageModel = llm
        self.template_values: dict[str, str] = {}

    @abstractmethod
    def compress_message_stack(self) -> None:
        raise NotImplementedError()

    def set_template_value(self, key: str, value: str) -> None:
        """Replaces a template key with the value provided when the chat template is
        applied."""
        self.template_values[key] = value

    def push_to_message_stack(
        self,
        message: BaseMessage,
    ) -> None:
        self.messages.append(message)

    def send_message(self, message: str) -> ChatResponse:
        """Sends a message to the AI model. Returns solution."""
        self.push_to_message_stack(message=HumanMessage(content=message))

        all_messages = list(self.ai_model_agent.system_messages.messages)
        all_messages.extend(self.messages)

        # Transform message stack to ChatPromptValue: If this is a ChatLLM then the
        # function will simply be an identity function that does nothing and simply
        # returns the messages as a ChatPromptValue. If this is a text generation
        # LLM, then the function should inject the config message around the
        # conversation to make the LLM behave like a ChatLLM.
        message_prompts: PromptValue = self.ai_model.apply_chat_template(
            messages=all_messages,
            **self.template_values,
        )

        response: ChatResponse
        try:
            result: LLMResult = self.llm.generate_prompt(
                prompts=[message_prompts],
            )

            response_message: BaseMessage = AIMessage(
                content=result.generations[0][0].text
            )

            self.push_to_message_stack(message=response_message)

            # Check if token limit has been exceeded.
            all_messages.append(response_message)
            new_tokens: int = self.llm.get_num_tokens_from_messages(
                messages=all_messages,
            )
            if new_tokens > self.ai_model.tokens:
                response = ChatResponse(
                    finish_reason=FinishReason.length,
                    message=response_message,
                    total_tokens=self.ai_model.tokens,
                )
            else:
                response = ChatResponse(
                    finish_reason=FinishReason.stop,
                    message=response_message,
                    total_tokens=new_tokens,
                )
        except Exception as e:
            print(f"There was an unkown error when generating a response: {e}")
            exit(1)

        return response



# Author: Yiannis Charalambous

from esbmc_ai.frontend.solution import apply_line_patch


def test_apply_line_patch() -> None:
    text = "\n".join(["a", "b", "c", "d", "e", "f", "g"])
    answer = "\n".join(["a", "b", "1", "g"])
    assert apply_line_patch(text, "1", 2, 5) == answer

    text = "\n".join(["a", "b", "c", "d", "e", "f", "g"])
    answer = "\n".join(["a", "b", "c", "1", "e", "f", "g"])
    assert apply_line_patch(text, "1", 3, 3) == answer
